library angular2.src.render.dom.shadow_dom.shadow_dom_compile_step;

import "../compiler/compile_step.dart" show CompileStep;
import "../compiler/compile_element.dart" show CompileElement;
import "../compiler/compile_control.dart" show CompileControl;
import "../../api.dart" show ViewDefinition;
import "shadow_dom_strategy.dart" show ShadowDomStrategy;
import "../util.dart" show NG_CONTENT_ELEMENT_NAME, isElementWithTag;

class ShadowDomCompileStep implements CompileStep {
  ShadowDomStrategy _shadowDomStrategy;
  ViewDefinition _view;
  ShadowDomCompileStep(this._shadowDomStrategy, this._view) {}
  process(
      CompileElement parent, CompileElement current, CompileControl control) {
    if (isElementWithTag(current.element, NG_CONTENT_ELEMENT_NAME)) {
      current.inheritedProtoView.bindNgContent();
    } else if (isElementWithTag(current.element, "style")) {
      this._processStyleElement(current, control);
    } else {
      var componentId =
          current.isBound() ? current.inheritedElementBinder.componentId : null;
      this._shadowDomStrategy.processElement(
          this._view.componentId, componentId, current.element);
    }
  }
  _processStyleElement(CompileElement current, CompileControl control) {
    this._shadowDomStrategy.processStyleElement(
        this._view.componentId, this._view.templateAbsUrl, current.element);
    // Style elements should not be further processed by the compiler, as they can not contain

    // bindings. Skipping further compiler steps allow speeding up the compilation process.
    control.ignoreCurrentElement();
  }
}
